/* ttt-args.c - Parse command-line arguments for ttt using getopt
 *
 * Copyright © 2005 Carl Worth
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Author: Carl Worth <cworth@cworth.org>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libgen.h>
#include <getopt.h>

#include "ttt-args.h"

static const char TTT_ARGS_PROGRAM_VERSION[] = VERSION;
static const char TTT_ARGS_PROGRAM_DESCRIPTION[] = "client-server tic-tac-toe game";
static const char TTT_ARGS_PROGRAM_BUG_ADDRESS[] = "<carl@theworths.org>";

static char TTT_ARGS_PROGRAM_ARGDOC[] = "";

/* XXX: getopt is rather annoying in that you must maintain a
 * string-encoding of the options in addition to the table. For
 * example, to enable the sample display option below, one would have
 * to add "d:" to the optstring.
 *
 * A useful exercise would be to write a function to generate the
 * optstring from the table.
 *
 * The other annoying thing is that the table does not include help
 * strings, so the args_help function below must also be maintained
 * manually.
 */

enum {
    TTT_ARGS_VAL_LOG_FILE = 256,
    TTT_ARGS_VAL_PID_FILE,
    TTT_ARGS_VAL_HELP,
    TTT_ARGS_VAL_VERSION,
};

static char ttt_args_optstring[] = "dh:p:";
static struct option ttt_args_options[] = {
    /* name,		has_arg,	flag,	val */
    {"host",		1,		0,	'h'},
    {"port",		1,		0,	'p'},
    {"detach",		0,		0,	'd'},
    {"log-file",	1,		0,	TTT_ARGS_VAL_LOG_FILE},
    {"pid-file",	1,		0,	TTT_ARGS_VAL_PID_FILE},
    {"help",		0,		0,	TTT_ARGS_VAL_HELP},
    {"version",		0,		0,	TTT_ARGS_VAL_VERSION},
    { 0 }
};

static void
ttt_args_help (const char *argv0)
{
    printf ("Usage: %s [OPTION] %s\n", argv0, TTT_ARGS_PROGRAM_ARGDOC);
    printf ("%s - %s\n", argv0, TTT_ARGS_PROGRAM_DESCRIPTION);
    puts ("");
    printf ("Options that are common to both client and server:\n");
    puts ("");
    printf ("  -h HOST, --host=HOST\tHost to connect/bind to [%s]\n",
	    TTT_ARGS_HOST_DEFAULT);
    printf ("  -p PORT, --port=PORT\tPort to connect/bind to [%s]\n",
	    TTT_ARGS_PORT_DEFAULT);
    printf ("           --help\tGive this help list\n");
    printf ("           --version\tPrint program version\n");
    puts ("");
    printf ("Options that are specific to the server:\n");
    puts ("");
    printf ("           --log-file=FILE\tFile to use for logging [stderr]\n");
    printf ("       -d, --detach\tDetach and daemonize\n");
    printf ("           --pid-file=FILE\tFile in which to save PID (if -d given)\n"
	    "                          \t[%s]\n",
	    TTT_ARGS_PID_FILE_DEFAULT);
}

#if 0
static void
ttt_args_usage (const char *argv0)
{
    printf ("Usage: %s [OPTION] %s\n", argv0, TTT_ARGS_PROGRAM_ARGDOC);
    printf ("Try `%s --help' for more information.\n", argv0);
}
#endif

int
ttt_args_parse(ttt_args_t *args, int argc, char *argv[])
{
    char *argv0_copy = strdup (argv[0]);
    char *argv0 = basename (argv0_copy);

    int c;

    args->host = TTT_ARGS_HOST_DEFAULT;
    args->port = TTT_ARGS_PORT_DEFAULT;
    args->log_file = TTT_ARGS_LOG_FILE_DEFAULT;
    args->detach = FALSE;
    args->pid_file = TTT_ARGS_PID_FILE_DEFAULT;

    while (1) {
	c = getopt_long (argc, argv, ttt_args_optstring, ttt_args_options, NULL);
	if (c == -1)
	    break;

	switch (c) {
	case 'h':
	    args->host = optarg;
	    break;
	case 'p':
	    args->port = optarg;
	    break;
	case TTT_ARGS_VAL_LOG_FILE:
	    args->log_file = optarg;
	    break;
	case 'd':
	    args->detach = TRUE;
	    break;
	case TTT_ARGS_VAL_PID_FILE:
	    args->pid_file = optarg;
	    break;
	case TTT_ARGS_VAL_VERSION:
	    printf ("%s\n", VERSION);
	    exit (0);
	    break;

	case TTT_ARGS_VAL_HELP:
	    ttt_args_help (argv0);
	    exit (0);
	    break;

	case '?':
	    ttt_args_help (argv0);
	    exit (1);
	    break;

	default:
	    fprintf (stderr, "Unhandled option `%c'\n", c);
	    break;
	}
    }

    free (argv0_copy);

    return 0;
}
