/* ttt-server.c - tic-tac-toe game server
 *
 * Copyright © 2005 Carl Worth
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Author: Carl Worth <cworth@cworth.org>
 */

#include "ttt-server.h"

#include "ttt-args.h"
#include "ttt-client.h"
#include "ttt-error.h"
#include "ttt-socket.h"

struct _ttt_server {
    pthread_mutex_t mutex;

    const char *host;
    const char *port;

    ttt_client_t **clients;
    int clients_size;
    int num_clients;
};

static void
ttt_server_init (ttt_server_t *server, const char *host, const char *port)
{
    pthread_mutex_init (&server->mutex, NULL);

    server->host = host;
    server->port = port;

    server->clients = NULL;
    server->clients_size = 0;
    server->num_clients = 0;
}

/* Exported: See ttt-server.h for documentation. */
ttt_error_t
ttt_server_register_client (ttt_server_t *server, ttt_client_t *client)
{
    int i;
    ttt_error_t error = TTT_ERROR_NONE;
    const char *username;

    pthread_mutex_lock (&server->mutex);

    username = ttt_client_get_username (client);

    assert (username != NULL);

    if (username[0] == '\0')
	return TTT_ERROR_INVALID_NAME;

    for (i = 0; i < server->num_clients; i++) {
	if (strcmp (ttt_client_get_username (server->clients[i]), username) == 0) {
	    error = TTT_ERROR_INVALID_NAME;
	    goto CLEANUP_LOCK;
	}
    }

    printf ("Client %s has joined.\r\n", username);

    server->num_clients++;

    if (server->num_clients > server->clients_size) {
	if (server->clients_size == 0)
	    server->clients_size = 1;
	else
	    server->clients_size *= 2;

	server->clients = xrealloc (server->clients,
				    server->clients_size * sizeof (ttt_client_t *));
    }

    server->clients [server->num_clients - 1] = client;

 CLEANUP_LOCK:
    pthread_mutex_unlock (&server->mutex);

    return error;
}

/* Exported: See ttt-server.h for documentation. */
void
ttt_server_unregister_client (ttt_server_t *server, ttt_client_t *client)
{
    int i;

    pthread_mutex_lock (&server->mutex);

    for (i = 0; i < server->num_clients; i++)
	if (server->clients[i] == client)
	    break;

    assert (i < server->num_clients);

    printf ("Client %s has left.\r\n", ttt_client_get_username (client));

    memmove (&server->clients[i], &server->clients[i+1],
	     (server->num_clients - i - 1) * sizeof (ttt_client_t *));

    server->num_clients--;

    pthread_mutex_unlock (&server->mutex);
}

/* Exported: See ttt-server.h for documentation. */
void
ttt_server_broadcast (ttt_server_t *server, const char *message)
{
    int i;

    pthread_mutex_lock (&server->mutex);

    for (i = 0; i < server->num_clients; i++)
	ttt_client_send (server->clients[i], message);

    pthread_mutex_unlock (&server->mutex);
}

/* Exported: See ttt-server.h for documentation. */
const char*
ttt_server_who (ttt_server_t *server)
{
    int i;
    char *response;

    pthread_mutex_lock (&server->mutex);

    xasprintf (&response, "WHO");

    for (i = 0; i < server->num_clients; i++)
	xasprintf (&response, "%s %s",
		   response,
		   ttt_client_get_username (server->clients[i]));

    xasprintf (&response, "%s\r\n", response);

    pthread_mutex_unlock (&server->mutex);

    return response;
}

/* Exported: See ttt-server.h for documentation. */
ttt_error_t
ttt_server_statistics (ttt_server_t *server, const char *username, char **response)
{
    ttt_bool_t usernamefound = FALSE;
    const char *client_username;
    int client_num_wins;
    int i;

    pthread_mutex_lock (&server->mutex);

    for (i = 0; i < server->num_clients; i++) {
	client_username = ttt_client_get_username (server->clients[i]);
	if (strcasecmp (username, client_username) == 0) {
	    usernamefound = TRUE;
	    client_num_wins = ttt_client_get_num_wins (server->clients[i]);
	    xasprintf (response, "STATISTICS %s \"\r\n"
		       "TICTACTOE WINS %d\r\n\"\r\n",
		       client_username,
		       client_num_wins);
	}
    }

    pthread_mutex_unlock (&server->mutex);

    if (!usernamefound)
	return TTT_ERROR_NO_USER;

    return TTT_ERROR_NONE;
}

/* Exported: See ttt-server.h for documentation. */
ttt_error_t
ttt_server_verify_username (ttt_server_t *server, const char *username)
{
    ttt_bool_t usernamefound = FALSE;
    const char *client_username;
    int i;

    pthread_mutex_lock (&server->mutex);

    for (i = 0; i < server->num_clients; i++) {
	client_username = ttt_client_get_username (server->clients[i]);
	if (strcasecmp (username, client_username) == 0)
	    usernamefound = TRUE;
    }

    pthread_mutex_unlock (&server->mutex);

    if (!usernamefound)
	return TTT_ERROR_NO_USER;

    return TTT_ERROR_NONE;
}

/* Exported: See ttt-server.h for documentation. */
const char*
ttt_server_get_host (ttt_server_t *server)
{
    return server->host;
}

/* Exported: See ttt-server.h for documentation. */
const char*
ttt_server_get_port (ttt_server_t *server)
{
    return server->port;
}

static const char *WELCOME_MESSAGE = 
"Welcome to ttt-server. The server is currently listening on:\r\n"
"\r\n"
"	%s:%s\r\n"
"\r\n"
"To test this, simply connect one or more clients to that host and port.\r\n"
"For example:\r\n"
"\r\n"
"	telnet %s %s\r\n"
"\r\n"
"The TTTP (tic-tac-toe protocol) has been partially implemented.\r\n"
"The following commands should work: HELO, HELP, MESSAGE, VERSION, QUIT, WHO.\r\n"
"\r\n";

static void
_ttt_server_accept (void *closure, int client_socket)
{
    ttt_server_t *server = closure;

    ttt_client_new (server, client_socket);
}

static void
_detach_and_write_child_pid_to (const char *filename)
{
    pid_t pid;

    /* Use the Unix double-fork trick to detach completely. See
     * setsid(2) for some details as to why two forks are
     * needed. */
    pid = xfork ();
    if (pid) {
	/* First parent just exits */
	exit (0);
    }
    
    chdir ("/");
    setsid ();
    
    pid = xfork ();
    if (pid) {
	/* Second parent exits after writing pid */
	FILE *file = xfopen (filename, "w");
	fprintf (file, "%d\n", pid);
	fclose (file);
	exit (0);
    }
    
    /* Final, detached child returns. */
}

int 
main (int argc, char **argv)
{
    ttt_args_t args;
    ttt_server_t server;
    int socket;

    ttt_args_parse (&args, argc, argv);

    if (args.log_file || args.detach) {
	FILE *log_file;
	/* In the detach case, we force redirection to a log file. */
	if (args.log_file == NULL)
	    args.log_file = "/var/log/ttt-server.log";
	log_file = fopen (args.log_file, "a");
	if (log_file == NULL) {
	    printf ("Warning: Failed to open log file %s: %s.\n",
		    args.log_file, strerror (errno));
	    printf ("Logging will be disabled.\n");
	    xdup2 (1, 2);
	} else {
	    xdup2 (fileno (log_file), 2);
	}
    }

    if (args.detach)
	_detach_and_write_child_pid_to (args.pid_file);

    socket = ttt_socket_create_server (args.host, args.port);

    if (args.detach) {
	xdup2 (2, 1);
	printf ("Server started listening on %s:%s\n", args.host, args.port);
    } else {
	printf (WELCOME_MESSAGE, args.host, args.port, args.host, args.port);
    }

    fclose (stdout);
    fclose (stdin);

    ttt_server_init (&server, args.host, args.port);

    while (1)
	ttt_socket_accept (socket, _ttt_server_accept, &server);

    /* We only reach here if something bad happened. */
    return 1;
}
